program ceiu;
uses crt;
{$I-}
const erTooParam=1;  {too many parameters}
      erFewParam=2;  {few parameters}
      erNoOption=3;  {no / option found}
      erNoParam=4;   {no param found}
      erNoFile=5;    {no file found}
      erNoRead=6;    {no file read}
      erNoCKTF100=7; {no CKTF 1.00 file found}
      erFLength=8;   {file length error}
      erNoMemory=9;  {out of memory}

      op_Nothing=0;
      op_I=1;        {option for generic info}
      op_B=2;        {option for type B info}
      op_T=3;        {option for test integrity}
      op_C=4;        {option for type C info}

      opTypeA=1;     {type A file}
      opTypeB=2;     {type B file}
      opTypeC=3;     {type C file}
      opTypeD=4;     {type D file}
      opTypeE=5;     {type E file}

      tyUnknown=0;
      tyCTTF100=1;   {Commodore Turbo Tape File 1.00}
      tyCKTF100=2;   {Commodore Kernal Tape File 1.00}
      tyCFTF100=3;   {Commodore Fast Tape File 1.00}
      tyCFTF101=4;   {Commodore Fast Tape File 1.01}
      tyCFTF102=5;   {Commodore Fast Tape File 1.02}
      tyCFTF103=6;   {Commodore Fast Tape File 1.03}
      tyCFTF104=7;   {Commodore Fast Tape File 1.04}

const HexDigits:array[0..15] of char='0123456789ABCDEF';

type Header4Byte=array[1..4] of byte;
     HexString=string[8];

type TBuffer=array[0..65534] of byte;
     PBuffer=^TBuffer;

var f:file of byte;
    Option:byte;        {contain option passed as parameter}
    FileName:string;
    Header:Header4Byte;
    FType:byte;         {contain the type of file}
    Buffer:PBuffer;     {a memory buffer to improve perfoemance in file read}

function HexByte(B:byte):HexString;
var Temp:HexString;
begin
  Temp[0]:=#2;
  Temp[1]:=HexDigits[B shr 4];
  Temp[2]:=HexDigits[B and $F];
  HexByte:=Temp;
end;

function HexWord(W:word):HexString;
var Temp:HexString;
begin
  Temp[0]:=#4;
  Temp[1]:=HexDigits[W shr 12];
  Temp[2]:=HexDigits[(W shr 8) and $F];
  Temp[3]:=HexDigits[(W shr 4) and $F];
  Temp[4]:=HexDigits[W and $F];
  HexWord:=Temp;
end;

procedure WriteError(Error:byte);
begin
  case Error of
   erTooParam:begin
                Writeln('There are too many parameters.');
              end;
   erFewParam:begin
                Writeln('You must specify 2 parameters.');
              end;
   erNoOption:begin
                Writeln('No /option are specify.');
              end;
   erNoParam:begin
               Writeln('C64 Extract Info Utility, Version 1.00');
               Writeln;
               Writeln('Usage: CEIU /options file_name');
               Writeln('/i, /I:  Extract generic information from input file');
               Writeln('/b, /B:  Extract the ending address from CKTF for Type B (used by READER3)');
               Writeln('/c, /C:  Extract starting address from CKTF for Type C (used by READER3)');
               Writeln('/t, /T:  Test integrity of file');
               Writeln;
             end;
   erNoFile:begin
              Writeln('Error opening input file');
            end;
   erNoRead:begin
              Writeln('Error reading input file');
            end;
   erNoCKTF100:begin
                 writeln('A CKTF ver 1.00 are needed for this option');
               end;
   erFLength:begin
               writeln('The length of the file data is different from that stored');
             end;
   erNoMemory:begin
                writeln('Out of memory');
              end;
  end;
  halt(Error);
end;

procedure OptionType(c:char);
begin
  case c of
   'i','I':Option:=op_I;
   'b','B':Option:=op_B;
   't','T':Option:=op_T;
   'c','C':Option:=op_C;
   else Option:=op_Nothing;
  end
end;

procedure ExtractInfo;  {analize command line to found parameters}
var s1,s2:string;
begin
  if ParamCount=0 then WriteError(erNoParam);
  if ParamCount>2 then WriteError(erTooParam);
  if ParamCount<2 then WriteError(erFewParam);
  s1:=ParamStr(1);
  if s1[1]='/' then OptionType(s1[2])
               else WriteError(erNoOption);
  FileName:=ParamStr(2);
end;

procedure OpenFile;
begin
  assign(f,FileName);
  reset(f);
  if IOResult<>0 then WriteError(erNoFile);
end;

procedure ReadAddr(pos:longint; var addr:word);
var b1,b2:byte;
begin
  seek(f,pos);
  read(f,b1);
  read(f,b2);
  if IOResult<>0 then WriteError(erNoRead);
  addr:=b1+b2*256;
end;

procedure ReadAddr_(pos:longint; var addr:word);
var b1,b2:byte;
begin
  seek(f,pos);
  read(f,b1);
  read(f,b2);
  if IOResult<>0 then WriteError(erNoRead);
  addr:=b2+b1*256;
end;

procedure ReadByte(pos:longint; var b:byte);
begin
  seek(f,pos);
  read(f,b);
  if IOResult<>0 then WriteError(erNoRead);
end;

function ReadName(pos:longint; len:byte):string;
var b:byte;
    i:byte;
begin
  seek(f,pos);
  for i:=1 to len do
   begin
     read(f,b);
     if IOResult<>0 then WriteError(erNoRead);
     readname[i]:=chr(b);
   end;
  ReadName[0]:=chr(len);
end;

procedure CTTF100;
var addr:word;
    b:byte;
begin
  writeln('Commodore 64 Turbo Tape File, Version 1.00');
  writeln;
  ReadByte(04,b);
  writeln('Header format:   ',HexByte(b));
  ReadByte(05,b);
  writeln('File format:     ',HexByte(b));
  ReadByte(06,b);
  writeln('Checksum:        ',HexByte(b));
  ReadAddr(07,addr);
  writeln('Start address:  ',HexWord(addr));
  ReadAddr(09,addr);
  writeln('End address:    ',HexWord(addr));
  ReadByte($0B,b);
  writeln('Null info:       ',HexByte(b));
  seek(f,$0C);
  read(f,b);
  if IOResult<>0 then WriteError(erNoRead);
  writeln('File name:      ',ReadName($0D,b));
  writeln;
end;

procedure CKTF100;
var b:byte;
    addr:word;
begin
  writeln('Commodore 64 Kernal Tape File, Version 1.00');
  writeln;
  ReadByte(04,b);
  writeln('Type of File:    ',HexByte(b));
  ReadAddr(05,addr);
  writeln('Start address:   ',HexWord(addr));
  ReadAddr(07,addr);
  writeln('End address:     ',HexWord(addr));
  writeln('File name:       ',ReadName(9,16));
  ReadByte($C0+04,b);
  writeln('Header Checksum: ',HexByte(b));
  ReadByte(FileSize(f)-1,b);
  writeln('File Checksum:   ',HexByte(b));
  writeln;
end;

procedure CFTF100;
var addr:word;
    b:byte;
begin
  writeln('Commodore 64 Fast Tape File, Version 1.00');
  writeln;
  ReadAddr(04,addr);
  writeln('Start address:  ',HexWord(addr));
  ReadAddr(06,addr);
  writeln('End address:    ',HexWord(addr));
  ReadByte(08,b);
  writeln('Checksum:        ',HexByte(b));
  writeln;
end;

procedure CFTF101;
var addr:word;
    b:byte;
begin
  writeln('Commodore 64 Fast Tape File, Version 1.01');
  writeln;
  ReadAddr(04,addr);
  writeln('Start address:  ',HexWord(addr));
  ReadAddr(06,addr);
  writeln('End address:    ',HexWord(addr));
  ReadByte(08,b);
  writeln('Checksum:        ',HexByte(b));
  writeln;
end;

procedure CFTF102;
begin
  writeln('Commodore 64 Fast Tape File, Version 1.02');
end;

procedure CFTF103;
var addr:word;
begin
  writeln('Commodore 64 Fast Tape File, Version 1.03');
  writeln;
  ReadAddr(04,addr);
  writeln('Start address:  ',HexWord(addr));
  ReadAddr(06,addr);
  writeln('End address:    ',HexWord(addr));
  ReadAddr(08,addr);
  writeln('Run address:    ',HexWord(addr));
  writeln;
end;

procedure CFTF104;
var addr1,addr2:word;
    b:byte;
    i:byte;
begin
  i:=0;
  writeln('Commodore 64 Fast Tape File, Version 1.04');
  writeln;
  seek(f,$04);
  while FilePos(f)<FileSize(f) do
   begin
     if ((i mod 5)=0) and (i<>0)
                    then begin
                           writeln('--- press any key ---');
                           repeat
                           until (keypressed);
                           b:=byte(readkey);
                           writeln;
                         end;
     i:=i+1;
     Read(f,b);
     writeln('Header bytes:   ',HexByte(b));
     if b>3 then seek(f,FilePos(f)+b-3);
     ReadAddr_(FilePos(f),addr2);
     ReadAddr_(FilePos(f),addr1);
     writeln('Start address:  ',HexWord(addr1));
     writeln('End address:    ',HexWord(addr2));
     writeln;
     if addr2-addr1=0 then inc(addr2);
     seek(f,longint(addr2)-longint(addr1)+FilePos(f))
   end;
end;

procedure FileInfo; {search for input file info}
begin
  case FType of
   tyCTTF100:CTTF100;
   tyCKTF100:CKTF100;
   tyCFTF100:CFTF100;
   tyCFTF101:CFTF101;
   tyCFTF102:CFTF102;
   tyCFTF103:CFTF103;
   tyCFTF104:CFTF104;
 end;
end;

procedure ExtractFType;
begin
  FType:=tyUnknown;
  read(f,header[1],header[2],header[3],header[4]);
  if IOResult<>0 then WriteError(erNoRead);
  if (header[1]=$02) and (header[2]=$13) and (header[3]=$13) and (header[4]=$05)
     then FType:=tyCTTF100;
  if (header[1]=$02) and (header[2]=$0A) and (header[3]=$13) and (header[4]=$05)
     then FType:=tyCKTF100;
  if (header[1]=$02) and (header[2]=$05) and (header[3]=$13) and (header[4]=$05)
     then FType:=tyCFTF100;
  if (header[1]=$02) and (header[2]=$05) and (header[3]=$13) and (header[4]=$25)
     then FType:=tyCFTF101;
  if (header[1]=$02) and (header[2]=$05) and (header[3]=$13) and (header[4]=$45)
     then FType:=tyCFTF102;
  if (header[1]=$02) and (header[2]=$05) and (header[3]=$13) and (header[4]=$65)
     then FType:=tyCFTF103;
  if (header[1]=$02) and (header[2]=$05) and (header[3]=$13) and (header[4]=$85)
     then FType:=tyCFTF104;
end;

function Check(var b:PBuffer; l:word):byte;
var Chk:byte;
    i:word;
begin
  Chk:=0;
  for i:=0 to l-1 do
      Chk:=Chk xor b^[i];
  Check:=Chk;
end;

procedure OptionB;
var b1,b2:byte;
begin
  if FType<>tyCKTF100 then WriteError(erNoCKTF100);
  FileInfo;
  writeln;
  seek(f,$030C-$02BA+$C5);
  read(f,b1);
  seek(f,$0313-$02BA+$C5);
  read(f,b2);
  if IOResult<>0 then WriteError(erNoRead);
  writeln('This information are to be uses only with READER3 type B file format');
  writeln;
  writeln('End Address:    ',HexWord(b2+b1*256));
end;

procedure OptionC;
var c:byte;
    i:byte;
begin
  if FType<>tyCKTF100 then WriteError(erNoCKTF100);
  GetMem(Buffer,sizeof(TBuffer));
  if Buffer=nil then WriteError(erNoMemory);
  FileInfo;
  seek(f,$114);
  read(f,c);
  seek(f,$18); {initial encripted area}
  BlockRead(file(f),buffer^,172);
  for i:=171 downto 0 do
    begin
      c:=c xor Buffer^[i];
      buffer^[i]:=c;
    end;
  writeln('Start Address: ',HexWord(Buffer^[67]+Buffer^[68]*256));
end;

procedure OptionT;
var SAddr:word;    {start addr}
    EAddr:word;    {end addr}
    Checksum:byte; {checksum that is strored}
    Chk:byte;      {chk calculated}
    FLength:word;  {file length}
    b:byte;
    d:word;
begin
  FileInfo;
  GetMem(Buffer,sizeof(TBuffer));
  if Buffer=nil then WriteError(erNoMemory);
  case FType of
   tyCTTF100:begin
               ReadByte($06,Checksum);
               ReadAddr($07,SAddr);
               ReadAddr($09,EAddr);
               ReadByte($0C,b);
               seek(f,$0D+b);
               FLength:=FileSize(f)-FilePos(f);
               if EAddr-SAddr<>FLength then WriteError(erFLength);
               BlockRead(file(f),Buffer^,FLength);
               Chk:=Check(Buffer,FLength);
               if Chk=Checksum then writeln('The Checksum and the file length are OK')
                               else writeln('The Checksum are INCORRECT');
             end;
   tyCKTF100:begin
               seek(f,$04);
               BlockRead(file(f),Buffer^,$C0+1);
               chk:=Check(Buffer,$C0+1);
               if Chk=0 then writeln('The Header Checksum is OK')
                        else writeln('The Header Checksum is INCORRECT');
               ReadAddr($05,SAddr);
               ReadAddr($07,EAddr);
               Seek(f,$C5);
               FLength:=FileSize(f)-FilePos(f);
               if EAddr-SAddr+1<>FLength then WriteError(erFLength);
               BlockRead(file(f),Buffer^,FLength);
               Chk:=Check(Buffer,FLength);
               if Chk=0 then writeln('The File Checksum and the file length are OK')
                        else writeln('The File Checksum is INCORRECT');
             end;
   tyCFTF100,
   tyCFTF101:begin
               ReadByte($08,Checksum);
               ReadAddr($04,SAddr);
               ReadAddr($06,EAddr);
               seek(f,$09);
               FLength:=FileSize(f)-FilePos(f);
               if EAddr-SAddr<>FLength then WriteError(erFLength);
               BlockRead(file(f),Buffer^,FLength);
               Chk:=Check(Buffer,FLength);
               if Chk=Checksum then writeln('The Checksum and the file length are OK')
                               else writeln('The Checksum are INCORRECT');
             end;
   tyCFTF103:begin
               ReadAddr($04,SAddr);
               ReadAddr($06,EAddr);
               seek(f,$0A);
               FLength:=FileSize(f)-FilePos(f);
               if EAddr-SAddr<>FLength then WriteError(erFLength);
               writeln('The file lenght is OK');
             end;
   tyCFTF104:begin
                SAddr:=0;
                seek(f,$04);
                while SAddr<>$03BB do
                 begin
                   ReadAddr(FilePos(f),SAddr);
                   ReadAddr(FilePos(f),EAddr);
                   d:=FilePos(f);
                   if SAddr<>$03A4 then seek(f,longint(EAddr)-longint(SAddr)+FilePos(f))
                    else begin
                          read(f,b);
                          seek(f,longint(EAddr)-longint(SAddr)+FilePos(f)+b-4);
                         end;
                 end;
                if FileSize(f)-D<>EAddr-SAddr then writeln('File Length is INCORRECT')
                                              else writeln('File Length is OK');
             end;
  end;
  FreeMem(Buffer,sizeof(TBuffer));
end;

begin
  clrscr;
  Writeln('C64 Extract Info Utility, Version 1.00');
  Writeln;
  ExtractInfo;
  OpenFile;
  ExtractFType;
  case Option of
   op_I:FileInfo;  {extract generic file info}
   op_B:OptionB;   {extract info for type B}
   op_T:OptionT;   {test integrity of file}
   op_C:OptionC;   {extract info for type C}
  end;
end.


25/01/98    35       : creazione procedure generali per paramentri
02/02/98  1:25       : parametro /i impostato
07/02/98    10       : /i per CKTF
11/02/98  1:40    10 : /t per CTTF e CKFT
14/11/98    20       : opzione /c e CFTF 103
15/11/98    15     5 : opzione /i e /t per CFTF 104
20/02/98    10       : \ modifiche
21/02/98     5       : \ modifiche per header bytes
          ----   ----
          4:40    15
