program make_cas;  {Make cassette for c64a emulator}
uses crt;
{$I-}

const erTooParam = 1;  {too many parameters}
      erNoParam  = 2;  {no param found}
      erOpening  = 3;  {no file found}
      erReading  = 4;  {no file read}
      erNoMemory = 5;  {out of memory}
      erWriting  = 6;  {no write to file}

      ceEndOfFile         =  1; {end of file unexpected}
      ceLineTooLong       =  2; {line too long}
      ceTypeExpected      =  3; {TYPE expected}
      ceEqualExpected     =  4; {= expected}
      ceCCDF100Expected   =  5; {#CCDF_1.00 expected}
      ceClosePExpected    =  6; {] expected}
      ceOpenPExpected     =  7; {[ expected}
      ceBlockExpected     =  8; {BLOCK expected}
      ceStringExpected    =  9; {"xx" expexted}
      ceImageExpected     = 10; {IMAGE expected}
      ceTextExpected      = 11; {TEXT expected}
      ceFileExpected      = 12; {FILE expected}
      ceTwoMExpected      = 13; {: expected}
      ceNameExpected      = 14; {NAME expected}
      cePatchExpected     = 15; {PATCH expected}
      ceLoadExpected      = 16; {LOAD expected}
      ceSaveExpected      = 17; {SAVE expected}
      ceVerifyExpected    = 18; {VERIFY expected}
      ceExeNumberExpected = 19; { $xxyy expected}
      ceSepExpected       = 20; {| expected}
      ceValidTypeExpected = 21; { #.. valid expected}
      ceLabelExpected     = 22; {LABEL expected}

type simbol=(siUnknown,      {unknown simbol}
             siCCDF100,
             siCTTF100,
             siCKTF100,
             siCFTF100,
             siCFTF101,
             siCFTF102,
             siCFTF103,
             siCFTF104,
             siTwoM,         {: simbol}
             siSep,          {| simbol}
             siString,       { "xx   " simbol}
             siExeNumber,    { $xxyy simbol}
             siEqual,        {= simbol}
             siOpenP,        {[xxx simbol}
             siCloseP,       {] simbol}
             siOpenPr,       {#xxx simbol}
             siText,         {[TEXT simbol}
             siType,         {[TYPE simbol}
             siImage,        {[IMAGE simbol}
             siBlock,        {[BLOCK simbol}
             siFile,         {[FILE simbol}
             siLoad,         {[LOAD simbol}
             siSave,         {[SAVE simbol}
             siVerify,       {[VERIFY simbol}
             siEnd,          {[END simbol}
             siName,         {[NAME simbol}
             siLabel,        {[LABEL simbol}
             siPatch         {[PATCH simbol}
             );

var InputFileName:string;
    OutputFileName:string;
    InputFile:text;
    OutputFile:file of byte;
    Row:word;                  {actual compile row}
    Col:word;                  {actual compile col. in line}
    sim:simbol;
    Number:word;               {number read as exe by parser}
    TextString:string;         {string read by parser}
    IsNewLine:boolean;

procedure CompileError(Error:byte);
begin
  write('(',Row,',',Col,') ');
  case Error of
    ceEndOfFile:Writeln('Unexpected end of file');
    ceLineTooLong:Writeln('Line too long');
    ceTypeExpected:Writeln('[TYPE] expected');
    ceEqualExpected:Writeln('= expected');
    ceCCDF100Expected:Writeln('#CCDF100 expected');
    ceClosePExpected:Writeln('] expected');
    ceOpenPExpected:Writeln('[ expected');
    ceBlockExpected:Writeln('[BLOCK] expected');
    ceStringExpected:Writeln('"..." expected');
    ceImageExpected:Writeln('[IMAGE] expected');
    ceTextExpected:Writeln('[TEXT] expected');
    ceFileExpected:Writeln('[FILE] expected');
    ceTwoMExpected:Writeln(': expected');
    ceNameExpected:Writeln('[NAME] expected');
    cePatchExpected:Writeln('[PATCH] expected');
    ceLoadExpected:Writeln('[LOAD] expected');
    ceSaveExpected:Writeln('[SAVE] expected');
    ceVerifyExpected:Writeln('[VERIFY] expected');
    ceExeNumberExpected:Writeln('$xyzk hexdecimal number expected');
    ceSepExpected:Writeln('| separator expected');
    ceValidTypeExpected:Writeln(' a valid type of file expected');
    ceLabelExpected:Writeln('[LABEL] expected');
  end;
  halt(Error);
end;

procedure WriteError(Error:byte);
begin
  case Error of
   erTooParam:begin
                Writeln('There are too many parameters.');
              end;
   erNoParam:begin
               Writeln('Usage: MAKE_CAS  input_file_name  output_file_name');
               Writeln;
             end;
   erOpening:Writeln('Error opening input file');
   erReading:Writeln('Error reading input file');
   erNoMemory:begin
                writeln('Out of memory');
              end;
  end;
  halt(Error);
end;

procedure SearchParam;
begin
  if (ParamCount<>2) then WriteError(erNoParam);
  InputFileName:=ParamStr(1);
  OutputFileName:=ParamStr(2);
end;

procedure OpenFile(var f:text; n:string);
begin
  assign(f,n);
  reset(f);
  if IOResult<>0 then WriteError(erOpening);
end;

procedure ReadLine(var f:text; var s:string);
begin
  Readln(f,s);
  if ioresult<>0 then WriteError(erReading);
end;

function ReadNextChar:char;
const s:string='';
      i:word=1;
var c:char;
begin
  repeat
    if s='' then begin
                   if eof(InputFile) then CompileError(ceEndOfFile);
                   ReadLine(InputFile,s);
                   Row:=Row+1;
                   Col:=1;
                   i:=1;
                   IsNewLine:=false;
                 end;
  until s<>'';
  c:=s[i];
  if s[0]=chr(i) then begin
                          s:='';
                          IsNewLine:=true;
                        end;
  Col:=Col+1;
  i:=i+1;
  if i=256 then CompileError(ceLineTooLong);
  ReadNextChar:=c;
end;

function ReadChar:char; {read char but not space}
var c:char;
begin
  repeat
    c:=ReadNextChar;
  until c<>' ';
  ReadChar:=c;
end;

procedure NewLine;  {read extra char to become in a new line}
var c:char;
begin
  repeat
    c:=ReadChar
  until IsNewLine
end;

function StringToExe(var s:string):boolean;
begin
  case chr(byte(s[4])) of
    '0'..'9':Number:=byte(s[4])-ord('0');
    'A'..'F':Number:=byte(s[4])-ord('A')+10;
    'a'..'f':Number:=byte(s[4])-ord('a')+10;
    else begin
           StringToExe:=false;
           exit
         end;
  end;
  case chr(byte(s[3])) of
    '0'..'9':Number:=Number+(byte(s[3])-ord('0'))*16;
    'A'..'F':Number:=Number+(byte(s[3])-ord('A')+10)*16;
    'a'..'f':Number:=Number+(byte(s[3])-ord('a')+10)*16;
    else begin
           StringToExe:=false;
           exit
         end;
  end;
  case chr(byte(s[2])) of
    '0'..'9':Number:=Number+(byte(s[2])-ord('0'))*256;
    'A'..'F':Number:=Number+(byte(s[2])-ord('A')+10)*256;
    'a'..'f':Number:=Number+(byte(s[2])-ord('a')+10)*256;
    else begin
           StringToExe:=false;
           exit
         end;
  end;
  case chr(byte(s[1])) of
    '0'..'9':Number:=Number+(byte(s[1])-ord('0'))*256*16;
    'A'..'F':Number:=Number+(byte(s[1])-ord('A')+10)*256*16;
    'a'..'f':Number:=Number+(byte(s[1])-ord('a')+10)*256*16;
    else begin
           StringToExe:=false;
           exit
         end;
  end;
  StringToExe:=True;
end;

procedure Parse;
var c:char;
    s:string;
    i:byte;
begin
  sim:=siUnknown;
  c:=ReadChar;
  case c of
   '=':sim:=siEqual;
   ':':sim:=siTwoM;
   '|':sim:=siSep;
   ']':sim:=siCloseP;
   '[':begin
         sim:=siOpenP;
         c:=ReadNextChar;
         case c of
          'T':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[3]:=ReadNextChar;
                s[0]:=#3;
                if s='YPE' then sim:=siType;
                if s='EXT' then sim:=siText;
              end;
          'I':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[3]:=ReadNextChar;
                s[4]:=ReadNextChar;
                s[0]:=#4;
                if s='MAGE' then sim:=siImage;
              end;
          'B':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[3]:=ReadNextChar;
                s[4]:=ReadNextChar;
                s[0]:=#4;
                if s='LOCK' then sim:=siBlock;
              end;
          'F':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[3]:=ReadNextChar;
                s[0]:=#3;
                if s='ILE' then sim:=siFile;
              end;
          'L':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[3]:=ReadNextChar;
                s[0]:=#3;
                if s='OAD' then sim:=siLoad;
                if s='ABE' then begin
                                  c:=ReadNextChar;
                                  if c='L' then sim:=siLabel
                                end;
              end;
          'V':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[3]:=ReadNextChar;
                s[4]:=ReadNextChar;
                s[5]:=ReadNextChar;
                s[0]:=#5;
                if s='ERIFY' then sim:=siVerify;
              end;
          'S':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[3]:=ReadNextChar;
                s[0]:=#3;
                if s='AVE' then sim:=siSave;
              end;
          'E':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[0]:=#2;
                if s='ND' then sim:=siEnd;
              end;
          'N':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[3]:=ReadNextChar;
                s[0]:=#3;
                if s='AME' then sim:=siName;
              end;
          'P':begin
                s[1]:=ReadNextChar;
                s[2]:=ReadNextChar;
                s[3]:=ReadNextChar;
                s[4]:=ReadNextChar;
                s[0]:=#4;
                if s='ATCH' then sim:=siPatch;
              end;
         end;
       end;
   '#':begin
         sim:=siOpenPr;
         s[1]:=ReadNextChar;
         s[2]:=ReadNextChar;
         s[3]:=ReadNextChar;
         s[4]:=ReadNextChar;
         s[5]:=ReadNextChar;
         s[6]:=ReadNextChar;
         s[7]:=ReadNextChar;
         s[8]:=ReadNextChar;
         s[9]:=ReadNextChar;
         s[0]:=#9;
         if s='CCDF_1.00' then sim:=siCCDF100;
         if s='CTTF_1.00' then sim:=siCTTF100;
         if s='CKTF_1.00' then sim:=siCKTF100;
         if s='CFTF_1.00' then sim:=siCFTF100;
         if s='CFTF_1.01' then sim:=siCFTF101;
         if s='CFTF_1.02' then sim:=siCFTF102;
         if s='CFTF_1.03' then sim:=siCFTF103;
         if s='CFTF_1.04' then sim:=siCFTF104;
       end;
   '$':begin
         s[1]:=ReadNextChar;
         s[2]:=ReadNextChar;
         s[3]:=ReadNextChar;
         s[4]:=ReadNextChar;
         s[0]:=#4;
         if StringToExe(s) then sim:=siExeNumber;
       end;
   '"':begin
         i:=0;
         repeat
           c:=ReadNextChar;
           i:=i+1;
           s[i]:=c;
        until (i<255) and (c='"') and (IsNewLine);
        s[0]:=chr(i-1);
        TextString:=s;
        if c='"' then sim:=siString
       end;
  end;
end;

procedure SearchClosePEqual;
begin
  Parse;
  if sim<>siCloseP then CompileError(ceClosePExpected);
  Parse;
  if sim<>siEqual then CompileError(ceEqualExpected);
end;

procedure SearchClosePTwoM;
begin
  Parse;
  if sim<>siCloseP then CompileError(ceClosePExpected);
  Parse;
  if sim<>siTwoM then CompileError(ceTwoMExpected);
end;

procedure SearchImage;
begin
  Parse;
  if sim<>siImage
   then if sim=siOpenP then CompileError(ceImageExpected)
                       else CompileError(ceOpenPExpected);
  SearchClosePEqual;
  Parse;
  if sim<>siString then CompileError(ceStringExpected);
end;

procedure SearchText;
begin
  Parse;
  if sim<>siText
   then if sim=siOpenP then CompileError(ceTextExpected)
                       else CompileError(ceOpenPExpected);
  SearchClosePEqual;
  Parse;
  if sim<>siString then CompileError(ceStringExpected);
end;

procedure SearchType;
begin
  Parse;
  if sim<>siType
    then if sim=siOpenP then CompileError(ceTypeExpected)
                        else CompileError(ceOpenPExpected);
  SearchClosePEqual;
end;

procedure SearchName;
begin
  Parse;
  if sim<>siName
    then if sim<>siOpenP then CompileError(ceNameExpected)
                         else CompileError(ceOpenPExpected);
  SearchClosePEqual;
  Parse;
  if sim<>siString then CompileError(ceStringExpected);
end;

procedure OutputWord; forward;

procedure SearchLoad;
begin
  Parse;
  if sim<>siLoad
    then if sim<>siOpenP then CompileError(ceLoadExpected)
                         else CompileError(ceOpenPExpected);
  SearchClosePEqual;
  Parse;
  if sim<>siExeNumber then CompileError(ceExeNumberExpected);
  OutputWord;
  if Number=0 then begin
                     OutputWord;
                     exit;
                   end;
  Parse;
  if sim<>siSep then CompileError(ceSepExpected);
  Parse;
  if sim<>siExeNumber then CompileError(ceExeNumberExpected);
  OutputWord;
end;

procedure SearchSave;
begin
  Parse;
  if sim<>siSave
    then if sim<>siOpenP then CompileError(ceSaveExpected)
                         else CompileError(ceOpenPExpected);
  SearchClosePEqual;
  Parse;
  if sim<>siExeNumber then CompileError(ceExeNumberExpected);
  OutputWord;
  if Number=0 then begin
                     OutputWord;
                     exit;
                   end;
  Parse;
  if sim<>siSep then CompileError(ceSepExpected);
  Parse;
  if sim<>siExeNumber then CompileError(ceExeNumberExpected);
  OutputWord;
end;

procedure SearchVerify;
begin
  Parse;
  if sim<>siVerify
    then if sim<>siOpenP then CompileError(ceVerifyExpected)
                         else CompileError(ceOpenPExpected);
  SearchClosePEqual;
  Parse;
  if sim<>siExeNumber then CompileError(ceExeNumberExpected);
  OutputWord;
  if Number=0 then begin
                     OutputWord;
                     exit;
                   end;
  Parse;
  if sim<>siSep then CompileError(ceSepExpected);
  Parse;
  if sim<>siExeNumber then CompileError(ceExeNumberExpected);
  OutputWord;
end;

procedure SearchPatch;
begin
  Parse;
  if sim<>siPatch
    then if sim<>siOpenP then CompileError(cePatchExpected)
                         else CompileError(ceOpenPExpected);
  SearchClosePTwoM;
  SearchLoad;
  SearchSave;
  SearchVerify;
end;

procedure SearchLabel;
begin
  Parse;
  if sim<>siLabel
    then if sim<>siOpenP then CompileError(ceLabelExpected)
                         else CompileError(ceOpenPExpected);
  SearchClosePEqual;
  Parse;
  if sim<>siString then CompileError(ceStringExpected);
end;

procedure OutputString(s:string);
var i:byte;
begin
  for i:=1 to byte(s[0]) do
    write(Outputfile,byte(s[i]));
  if ioresult<>0 then WriteError(erWriting);
end;

procedure OutputString12(s:string);
var i:byte;
    b:byte;
begin
  if s='' then begin
                 OutputString12('*           ');
                 exit;
               end;
  if byte(s[0])>12 then s[0]:=#12;
  for i:=1 to byte(s[0]) do
    write(OutputFile,byte(s[i]));
  b:=byte(' ');
  for i:=byte(s[0])+1 to 12 do
    write(OutputFile,b);
  if ioresult<>0 then WriteError(erWriting);
end;

procedure OutputString16(s:string);
var i:byte;
    b:byte;
begin
  if byte(s[0])>16 then s[0]:=#16;
  for i:=1 to byte(s[0]) do
    write(Outputfile,byte(s[i]));
  b:=byte(' ');
  for i:=byte(s[0])+1 to 16 do
    write(OutputFile,b);
  if ioresult<>0 then WriteError(erWriting);
end;

procedure OutputString30(s:string);
var i:byte;
    b:byte;
begin
  if byte(s[0])>30 then s[0]:=#30;
  for i:=1 to byte(s[0]) do
    write(Outputfile,byte(s[i]));
  b:=byte(' ');
  for i:=byte(s[0])+1 to 30 do
    write(OutputFile,b);
  if ioresult<>0 then WriteError(erWriting);
end;

procedure OutputType;
begin
  case sim of
    siCTTF100:OutputString(#2+#19+#19+#5);
    siCKTF100:OutputString(#2+#10+#19+#5);
    siCFTF100:OutputString(#2+#5+#19+#5);
    siCFTF101:OutputString(#2+#5+#19+#37);
    siCFTF102:OutputString(#2+#5+#19+#69);
    siCFTF103:OutputString(#2+#5+#19+#101);
    siCFTF104:OutputString(#2+#5+#19+#133);
  end;
end;

procedure OutputWord;
var b1,b2:byte;
begin
  b1:=lo(Number);
  b2:=hi(Number);
  write(OutputFile,b1);
  write(OutputFile,b2);
  if ioresult<>0 then WriteError(erWriting);
end;

procedure Compile;
begin
  SearchType;
  Parse;
  if sim<>siCCDF100 then CompileError(ceCCDF100Expected);
  OutputString(#2+#2+#3+#5);
  SearchImage;
  OutputString12(TextString);
  SearchText;
  OutputString12(TextString);
  SearchLabel;
  OutputString30(TextString);
  Parse;
  while sim<>siEnd do
    begin
      if sim<>siBlock
        then if sim=siOpenP then CompileError(ceBlockExpected)
                            else CompileError(ceOpenPExpected);
      SearchClosePEqual;
      OutputString('>'); {another block}
      Parse;
      if sim<>siString then CompileError(ceStringExpected);
      OutputString16(TextString);
      SearchImage;
      OutputString12(TextString);
      SearchText;
      OutputString12(TextString);
      Parse;
      while (sim<>siBlock) and (sim<>siEnd) do
        begin
          if sim<>siFile
           then if sim=siOpenP then CompileError(ceFileExpected)
                               else CompileError(ceOpenPExpected);
          OutputString('');
          SearchClosePTwoM;
          SearchName;
          OutputString12(TextString);
          SearchType;
          Parse;
          if not(sim in
            [siCTTF100,
             siCKTF100,
             siCFTF100,
             siCFTF101,
             siCFTF102,
             siCFTF103,
             siCFTF104])
            then CompileError(ceValidTypeExpected);
          OutputType;
          SearchPatch;
          Parse;
        end;
      OutputString('!');
    end;
  OutputString('^'); {end block}
end;

begin
  Row:=1;
  clrscr;
  Writeln('Make Cassette Utility, Version 1.00');
  Writeln;
  SearchParam;
  OpenFile(InputFile,InputFileName);
  assign(OutputFile,OutputfileName);
  rewrite(Outputfile);
  Compile;
  close(InputFile);
  close(Outputfile);
end.

example:

[TYPE]=  #CCDF_1.00
[IMAGE]= ""
[TEXT]=  "TESTO.TXT"
[LABEL]= "TESTO DESCRIZIONE CASSETTA"

[BLOCK]=    "CASSETTA DI PROVA"
  [IMAGE]=  "GAME1.GIF"
  [TEXT]=   ""
  [FILE]:
    [NAME]= "GAME1.T@0"
    [TYPE]= #CTTF_1.00
    [PATCH]:
      [LOAD]=   $xxyy | $jjll
      [SAVE]=   $0000
      [VERIFY]= $0000
  ...
  [FILE]:
     ...
[BLOCK]:
  ...
[END]

Low level:

 00h      02h    : Commodore
 01h      02h    : Cassette
 02h      03h    : Definition
 03h      05h    : File (ver 1.00)
 04h,0Ch  IMAGE  : name of image (8.3 format) or '*           ' for nothing
 ::
 10h,0Ch  TEXT   : name of text (8.3 format) or '*         ' for nothing
 ::
 1Ch,1Eh  LABEL  : label comment (30 char length)
 ::
 3Ah      BLOCK  : '>'=another, '^'=nothing ather
 $$,10h          : block name comment (16 char length)
 ::
 $$,0Ch   IMAGE  : name of image (8.3 format) or '*           ' for nothing
 ::
 $$,0Ch   TEXT   : name of text (8.3 format) or '*         ' for nothing
 ::
 $$,01h   FILE   : ''=another, '!'=nothing ather
 $$,0Ch   NAME   : name of file (8.3 format)
 ::
 $$,04h   TYPE   : type of file (header 4 format)
 ::
 $$,04h   LOAD   : 2 word number
 $$,04h   SAVE   : 2 word number
 $$,04h   VERIFY : 2 word number
 ...

10/03/98  1:05     : write error, parser
12/03/98  1:10   5 : parser+ compiler
15/03/98    25     : compiler
30/03/98    30     : compiler, schema quasi completo
19/04/98    15  30 : output code + debug per compiler
21/04/98  1:00     : output code completato
          .....  ..
          4:35  35


